/* codeconv.c */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#undef USE_ICONV

#include <glib.h>
#include <string.h>
#include <ctype.h>
#ifdef USE_ICONV
#  include <iconv.h>
#endif /* USE_ICONV */
#include <kcc.h>

#include "intl.h"
#include "codeconv.h"
#include "utils.h"

#define iskanji(c) \
	(((c) & 0xff) >= 0xa1 && ((c) & 0xff) <= 0xfe)
#define iseucss(c) \
	(((c) & 0xff) == 0x8e || ((c) & 0xff) == 0x8f)
#define isunprintablekanji(c) \
	(((c) & 0xff) >= 0xa9 && ((c) & 0xff) <= 0xaf)

void conv_jistoeuc(gchar *outbuf, gint outlen, const gchar *inbuf)
{
#ifdef USE_ICONV
	static iconv_t cd;
	size_t inlen;

	if (!cd)
		cd = iconv_open("EUC-JP", OUT_CHARSET);

	inlen = strlen(inbuf) + 1;
	memset(outbuf, 0, inlen);
	iconv(cd, &inbuf, &inlen, &outbuf, &outlen);
#else
	KCC_filter(outbuf, "EUC", (gchar *)inbuf, "JISBB", 0, 0, 0);
#endif /* USE_ICONV */
}

void conv_euctojis(gchar *outbuf, gint outlen, const gchar *inbuf)
{
#ifdef USE_ICONV
	static iconv_t cd;
	size_t inlen;

	if (!cd)
		cd = iconv_open(OUT_CHARSET, "EUC-JP");

	inlen = strlen(inbuf) + 1;
	memset(outbuf, 0, MIN(inlen * 3, outlen));
	iconv(cd, &inbuf, &inlen, &outbuf, &outlen);
#else
	size_t inlen, len;

	inlen = strlen(inbuf);
	if (iskanji(inbuf[inlen - 1]) || iseucss(inbuf[inlen - 1])) {
		/* if tail end of the string is not ended with ascii,
		   add dummy return code. */
		gchar *tmpin, *tmpout;

		/* length of original string + '\n' + '\0' */
		tmpin = alloca(inlen + 2);
		if (tmpin == NULL) {
			g_warning(_("can't allocate memory\n"));
			KCC_filter(outbuf, "JISBB", (gchar *)inbuf, "EUC",
				   0, 0, 0);
			return;
		}
		strcpy(tmpin, inbuf);
		tmpin[inlen] = '\n';
		tmpin[inlen + 1] = '\0';

		tmpout = alloca(outlen + 1);
		if (tmpout == NULL) {
			g_warning(_("can't allocate memory\n"));
			KCC_filter(outbuf, "JISBB", (gchar *)inbuf, "EUC",
				   0, 0, 0);
			return;
		}

		KCC_filter(tmpout, "JISBB", tmpin, "EUC", 0, 0, 0);
		len = strlen(tmpout);
		if (tmpout[len - 1] == '\n')
			tmpout[len - 1] = '\0';
		strncpy2(outbuf, tmpout, outlen);
	} else
		KCC_filter(outbuf, "JISBB", (gchar *)inbuf, "EUC", 0, 0, 0);
#endif /* USE_ICONV */
}

void conv_sjistoeuc(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	KCC_filter(outbuf, "EUC", (gchar *)inbuf, "SJIS", 0, 0, 0);
}

void conv_anytoeuc(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	KCC_filter(outbuf, "EUC", (gchar *)inbuf, "AUTO", 0, 0, 0);
}

void conv_anytojis(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	KCC_filter(outbuf, "JISBB", (gchar *)inbuf, "AUTO", 0, 0, 0);
}

#define SUBST_CHAR	'_'

void conv_unreadable_eucjp(gchar *str)
{
	register guchar *p = str;

	while (*p != '\0') {
		if (isascii(*p)) {
			/* convert CR+LF -> LF */
			if (*p == '\r' && *(p + 1) == '\n')
				memmove(p, p + 1, strlen(p));
			/* printable 7 bit code */
			p++;
		} else if (iskanji(*p)) {
			if (iskanji(*(p + 1)) && !isunprintablekanji(*p))
				/* printable euc-jp code */
				p += 2;
			else {
				/* substitute unprintable code */
				*p++ = SUBST_CHAR;
				if (*p != '\0') {
					if (isascii(*p))
						p++;
					else
						*p++ = SUBST_CHAR;
				}
			}
		} else if (iseucss(*p)) {
			if ((*(p + 1) & 0x80) != 0)
				/* euc-jp hankaku kana */
				p += 2;
			else
				*p++ = SUBST_CHAR;
		} else
			/* substitute unprintable 1 byte code */
			*p++ = SUBST_CHAR;
	}
}

void conv_unreadable_8bit(gchar *str)
{
	register guchar *p = str;

	while (*p != '\0') {
		/* convert CR+LF -> LF */
		if (*p == '\r' && *(p + 1) == '\n')
			memmove(p, p + 1, strlen(p));
		else if (!isascii(*p)) *p = SUBST_CHAR;
		p++;
	}
}

#define NCV	'\0'

void conv_mb_alnum(gchar *str)
{
	static guchar char_tbl[] = {
		/* 0xa0 - 0xaf */
		NCV, ' ', NCV, NCV, ',', '.', NCV, ':',
		';', '?', '!', NCV, NCV, NCV, NCV, NCV,
		/* 0xb0 - 0xbf */
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV,
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV,
		/* 0xc0 - 0xcf */
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV,
		NCV, NCV, '(', ')', NCV, NCV, '[', ']',
		/* 0xd0 - 0xdf */
		'{', '}', NCV, NCV, NCV, NCV, NCV, NCV,
		NCV, NCV, NCV, NCV, '+', '-', NCV, NCV,
		/* 0xe0 - 0xef */
		NCV, '=', NCV, '<', '>', NCV, NCV, NCV,
		NCV, NCV, NCV, NCV, NCV, NCV, NCV, NCV
	};

	register guchar *p = str;
	register gint len;

	len = strlen(str);

	while (len > 1) {
		if (*p == 0xa3) {
			register guchar ch = *(p + 1);

			if (ch >= 0xb0 && ch <= 0xfa) {
				/* [a-zA-Z] */
				*p = ch & 0x7f;
				p++;
				len--;
				memmove(p, p + 1, len);
				len--;
			} else  {
				p += 2;
				len -= 2;
			}
		} else if (*p == 0xa1) {
			register guchar ch = *(p + 1);

			if (ch >= 0xa0 && ch <= 0xef &&
			    NCV != char_tbl[ch - 0xa0]) {
				*p = char_tbl[ch - 0xa0];
				p++;
				len--;
				memmove(p, p + 1, len);
				len--;
			} else {
				p += 2;
				len -= 2;
			}
		} else if (iskanji(*p)) {
			p += 2;
			len -= 2;
		} else {
			p++;
			len--;
		}
	}
}

void conv_jistodisp(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	conv_jistoeuc(outbuf, outlen, inbuf);
	conv_unreadable_eucjp(outbuf);
}

void conv_sjistodisp(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	conv_sjistoeuc(outbuf, outlen, inbuf);
	conv_unreadable_eucjp(outbuf);
}

void conv_euctodisp(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	strncpy2(outbuf, inbuf, outlen);
	conv_unreadable_eucjp(outbuf);
}

void conv_ustodisp(gchar *outbuf, gint outlen, const gchar *inbuf)
{
	strncpy2(outbuf, inbuf, outlen);
	conv_unreadable_8bit(outbuf);
}
